@@ -17,22 +17,28 @@ eigeningenuity.core) or a full URL to a JSON Bridge Historian instance.
       print(asctime(time.gmtime(dp['timestamp'])), dp['value'])
 """
 
-import requests, warnings
-from urllib.parse import quote as urlquote
+import requests
 from typing import Union
 from datetime import datetime
 
 from requests.exceptions import ConnectionError
 from urllib.error import URLError
 from eigeningenuity.core import get_default_server, EigenServer
-from eigeningenuity.util import force_list, time_to_epoch_millis, EigenException, \
-        multi_response_to_df, merge_tags_from_response, update_keys, _do_historian_multi_request, format_output
+from eigeningenuity.util import (
+    force_list,
+    time_to_epoch_millis,
+    EigenException,
+    merge_tags_from_response,
+    update_keys,
+    _do_historian_multi_request,
+    format_output,
+)
 
-class HistorianMulti (object):
-    """A Historian which talks the Eigen Historian Json Bridge protocol.
-    """
 
-    def __init__(self, baseurl, auth, historian):
+class HistorianMulti(object):
+    """A Historian which talks the Eigen Historian Json Bridge protocol."""
+
+    def __init__(self, baseurl, auth, historian, timestamp_format):
         """This is a constructor. It takes in a URL like http://infra:8080/historian-servlet/jsonbridge/Demo-influxdb"""
         self.baseurl = baseurl
         self.serverurl = baseurl + "multi/"
@@ -45,46 +51,50 @@ class HistorianMulti (object):
         self.historian = ""
         if historian:
             self.historian = historian + "/"
-        
-        
+        self.timestamp_format = timestamp_format
+
     def _testConnection(self):
         try:
             status = requests.get(self.serverurl, verify=False).status_code
             if status != 200:
                 raise ConnectionError(
-                    "Invalid API Response from " + self.serverurl + ". Please check the url is correct and the instance is up.")
+                    "Invalid API Response from "
+                    + self.serverurl
+                    + ". Please check the url is correct and the instance is up."
+                )
         except (URLError, ConnectionError):
-            raise ConnectionError ("Failed to connect to ingenuity instance at " + self.serverurl + ". Please check the url is correct and the instance is up.")
+            raise ConnectionError(
+                "Failed to connect to ingenuity instance at "
+                + self.serverurl
+                + ". Please check the url is correct and the instance is up."
+            )
 
-   
-   #SECTION Read Requests
-   
-    def getCurrentDataPoints(self,tags: Union[str, list], output: str = "json"):
+    # SECTION Read Requests
+
+    def getCurrentDataPoints(self, tags: Union[str, list], output: str = "json"):
         tags = force_list(tags)
-        body = {
-            "requests": {}
-        }
+        body = {"requests": {}}
 
         for tag in tags:
             fulltag = tag
             if "/" not in tag:
                 fulltag = self.historian + tag
-            request = {
-                "type": "CURRENT_POINT",
-                "tag": fulltag,
-                "details": {}
-            }
+            request = {"type": "CURRENT_POINT", "tag": fulltag, "details": {}}
 
             body["requests"][tag] = request
 
-        response = _do_historian_multi_request(self.serverurl,body,self.auth)
+        response = _do_historian_multi_request(self.serverurl, body, self.auth)
 
-        return format_output(output,response,tags)
+        return format_output(
+            output, response, tags, timestamp_format=self.timestamp_format
+        )
 
-
-
-    def getInterpolatedPoints(self, tags: Union[str, list], timestamps: Union[list, int, float, str, datetime],
-                              output: str = "json"):
+    def getInterpolatedPoints(
+        self,
+        tags: Union[str, list],
+        timestamps: Union[list, int, float, str, datetime],
+        output: str = "json",
+    ):
         """
         Return a specified number of interpolated datapoints between a start and end timestamp for one or more datatags
 
@@ -102,9 +112,7 @@ class HistorianMulti (object):
             RemoteServerException: When the provided historian could not be found, or the API response contains an error
             RunTimeError/HTTPError: No response from the API
         """
-        body = {
-            "requests": {}
-        }
+        body = {"requests": {}}
 
         tags = force_list(tags)
         timestamps = force_list(timestamps)
@@ -118,22 +126,26 @@ class HistorianMulti (object):
                 request = {
                     "type": "HISTORICAL_POINT",
                     "tag": fulltag,
-                    "details": {
-                        "at": timestamp
-                    }
+                    "details": {"at": timestamp},
                 }
 
                 body["requests"][tag + "-" + str(index)] = request
 
-
-        response = _do_historian_multi_request(self.serverurl,body,self.auth)
+        response = _do_historian_multi_request(self.serverurl, body, self.auth)
         items = merge_tags_from_response(response["results"])
 
-        return format_output(output,items,tags,timestamps)
+        return format_output(
+            output, items, tags, timestamps, timestamp_format=self.timestamp_format
+        )
 
-
-    def getInterpolatedRange(self, tags: Union[str, list], start: Union[int, float, str, datetime],
-                             end: Union[int, float, str, datetime], count: int = 1000, output: str = "json"):
+    def getInterpolatedRange(
+        self,
+        tags: Union[str, list],
+        start: Union[int, float, str, datetime],
+        end: Union[int, float, str, datetime],
+        count: int = 1000,
+        output: str = "json",
+    ):
         """Return a specified number of interpolated datapoints between a start and end timestamp for one or more datatags
 
         Args:
@@ -153,9 +165,7 @@ class HistorianMulti (object):
             RunTimeError/HTTPError: No response from the API
         """
 
-        body = {
-            "requests": {}
-        }
+        body = {"requests": {}}
 
         tags = force_list(tags)
 
@@ -169,18 +179,26 @@ class HistorianMulti (object):
                 "details": {
                     "from": time_to_epoch_millis(start),
                     "to": time_to_epoch_millis(end),
-                    "count": count
-                }
+                    "count": count,
+                },
             }
-            
+
             body["requests"][tag] = request
 
-        response = _do_historian_multi_request(self.serverurl,body,self.auth)
+        response = _do_historian_multi_request(self.serverurl, body, self.auth)
 
-        return format_output(output,response,tags)
-    
-    def getRawDatapoints(self, tags: Union[str, list], start: Union[int, float, str, datetime],
-                             end: Union[int, float, str, datetime], maxpoints: int = 1000, output: str = "json"):
+        return format_output(
+            output, response, tags, timestamp_format=self.timestamp_format
+        )
+
+    def getRawDatapoints(
+        self,
+        tags: Union[str, list],
+        start: Union[int, float, str, datetime],
+        end: Union[int, float, str, datetime],
+        maxpoints: int = 1000,
+        output: str = "json",
+    ):
         """Return up to a specified number of raw datapoints between a start and end timestamp for one or more datatags
 
         Args:
@@ -200,9 +218,7 @@ class HistorianMulti (object):
             RunTimeError/HTTPError: No response from the API
         """
 
-        body = {
-            "requests": {}
-        }
+        body = {"requests": {}}
 
         tags = force_list(tags)
 
@@ -216,18 +232,25 @@ class HistorianMulti (object):
                 "details": {
                     "from": time_to_epoch_millis(start),
                     "to": time_to_epoch_millis(end),
-                    "maxPoints": maxpoints
-                }
+                    "maxPoints": maxpoints,
+                },
             }
-            
+
             body["requests"][tag] = request
 
-        response = _do_historian_multi_request(self.serverurl,body,self.auth)
+        response = _do_historian_multi_request(self.serverurl, body, self.auth)
 
-        return format_output(output,response,tags)
-    
-    def getClosestRawPoint(self, tags: Union[str, list], timestamps: Union[int, float, str, datetime, list],
-                             before_or_after:str="AFTER_OR_AT", output: str = "json"):
+        return format_output(
+            output, response, tags, timestamp_format=self.timestamp_format
+        )
+
+    def getClosestRawPoint(
+        self,
+        tags: Union[str, list],
+        timestamps: Union[int, float, str, datetime, list],
+        before_or_after: str = "AFTER_OR_AT",
+        output: str = "json",
+    ):
         """Return up to a specified number of raw datapoints between a start and end timestamp for one or more datatags
 
         Args:
@@ -248,12 +271,17 @@ class HistorianMulti (object):
         """
         before_or_after = before_or_after.upper()
 
-        if before_or_after.upper() not in ["BEFORE", "BEFORE_OR_AT", "AFTER", "AFTER_OR_AT"]:
-            raise EigenException('before_or_after must be one of ["BEFORE", "BEFORE_OR_AT", "AFTER", "AFTER_OR_AT"]')
+        if before_or_after.upper() not in [
+            "BEFORE",
+            "BEFORE_OR_AT",
+            "AFTER",
+            "AFTER_OR_AT",
+        ]:
+            raise EigenException(
+                'before_or_after must be one of ["BEFORE", "BEFORE_OR_AT", "AFTER", "AFTER_OR_AT"]'
+            )
 
-        body = {
-            "requests": {}
-        }
+        body = {"requests": {}}
 
         tags = force_list(tags)
         timestamps = force_list(timestamps)
@@ -269,18 +297,19 @@ class HistorianMulti (object):
                     "tag": fulltag,
                     "details": {
                         "at": time_to_epoch_millis(timestamp),
-                    }
+                    },
                 }
-            
+
                 body["requests"][tag + "-" + str(index)] = request
 
-
-        response = _do_historian_multi_request(self.serverurl,body,self.auth)
+        response = _do_historian_multi_request(self.serverurl, body, self.auth)
         items = merge_tags_from_response(response["results"])
 
-        return format_output(output,items,tags,timestamps)
-    
-    def listDataTags(self, historian:str=None, limit:int=100, match:str=""):
+        return format_output(
+            output, items, tags, timestamps, timestamp_format=self.timestamp_format
+        )
+
+    def listDataTags(self, historian: str = None, limit: int = 100, match: str = ""):
         """List all tags in a historian, or those matching a wildcard
 
         Args:
@@ -301,8 +330,8 @@ class HistorianMulti (object):
         tags = [tag["tagName"] for tag in metadata]
 
         return tags
-    
-    def getMetaData(self, historian:str=None, limit:int=100, match:str=""):
+
+    def getMetaData(self, historian: str = None, limit: int = 100, match: str = ""):
         """Get Metadata for all tags in a historian, or those matching a wildcard
 
         Args:
@@ -321,12 +350,17 @@ class HistorianMulti (object):
         if historian is None:
             historian = self.historian
 
-        resp = _do_historian_multi_request(self.searchurl + f'?historian={historian}&limit={limit}&search={match}',{},self.auth,get_request=True)
+        resp = _do_historian_multi_request(
+            self.searchurl + f"?historian={historian}&limit={limit}&search={match}",
+            {},
+            self.auth,
+            get_request=True,
+        )
         metadata = resp["tags"]
 
         return metadata
-    
-    def listHistorians(self, historian:str=None, limit:int=100, match:str=""):
+
+    def listHistorians(self, historian: str = None, limit: int = 100, match: str = ""):
         """Get Metadata for all tags in a historian, or those matching a wildcard
 
         Args:
@@ -345,14 +379,26 @@ class HistorianMulti (object):
         if historian is None:
             historian = self.historian
 
-        resp = _do_historian_multi_request(self.searchurl + f'?historian={historian}&limit={limit}&search={match}',{},self.auth,get_request=True)
+        resp = _do_historian_multi_request(
+            self.searchurl + f"?historian={historian}&limit={limit}&search={match}",
+            {},
+            self.auth,
+            get_request=True,
+        )
         metadata = resp["tags"]
 
         return metadata
-    
-    #SECTION - Tag Metadata Requests
-    
-    def createTag(self, tag:str, description:str=None, units:str=None, stepped:bool=False, update_existing:bool=False):
+
+    # SECTION - Tag Metadata Requests
+
+    def createTag(
+        self,
+        tag: str,
+        description: str = None,
+        units: str = None,
+        stepped: bool = False,
+        update_existing: bool = False,
+    ):
         """Create a Tag With Metadata - Writing points to a new tag directly will also create it, but will not create metadata
 
         Args:
@@ -377,27 +423,33 @@ class HistorianMulti (object):
 
         metadata = {}
 
-        for key, value in [("description", description), ("units", units), ("stepped", stepped)]:
+        for key, value in [
+            ("description", description),
+            ("units", units),
+            ("stepped", stepped),
+        ]:
             if value is not None:
                 metadata[key] = value
 
-        body = {
-            "entries": {
-                f"{tag}": metadata
-            },
-            "updateIfExists": update_existing
-        }
+        body = {"entries": {f"{tag}": metadata}, "updateIfExists": update_existing}
+
+        resp = _do_historian_multi_request(url, body, self.auth)
 
-        resp = _do_historian_multi_request(url,body,self.auth)
-        
         if len(resp["errors"].keys()):
             success = False
         else:
             success = True
 
         return {"success": success, "errors": resp["errors"]}
-    
-    def updateTag(self, tag:str, description:str=None, units:str=None, stepped:bool=False, create_missing:bool=False):
+
+    def updateTag(
+        self,
+        tag: str,
+        description: str = None,
+        units: str = None,
+        stepped: bool = False,
+        create_missing: bool = False,
+    ):
         """Update or Create Metadata for an existing tag
 
         Args:
@@ -422,18 +474,17 @@ class HistorianMulti (object):
 
         metadata = {}
 
-        for key, value in [("description", description), ("units", units), ("stepped", stepped)]:
+        for key, value in [
+            ("description", description),
+            ("units", units),
+            ("stepped", stepped),
+        ]:
             if value is not None:
                 metadata[key] = value
 
-        body = {
-            "entries": {
-                f"{tag}": metadata
-            },
-            "createIfMissing": create_missing
-        }
+        body = {"entries": {f"{tag}": metadata}, "createIfMissing": create_missing}
 
-        resp = _do_historian_multi_request(url,body,self.auth)
+        resp = _do_historian_multi_request(url, body, self.auth)
 
         if len(resp["errors"].keys()):
             success = False
@@ -441,8 +492,16 @@ class HistorianMulti (object):
             success = True
 
         return {"success": success, "errors": resp["errors"]}
-    
-    def createAndPopulateTag(self, tag:str, points: Union[dict, list], description:str=None, units:str=None, stepped:bool=False, update_existing:bool=True):
+
+    def createAndPopulateTag(
+        self,
+        tag: str,
+        points: Union[dict, list],
+        description: str = None,
+        units: str = None,
+        stepped: bool = False,
+        update_existing: bool = True,
+    ):
         """Create a Tag With Metadata, and batch import points in a single function
 
         Args:
@@ -468,45 +527,41 @@ class HistorianMulti (object):
 
         metadata = {}
 
-        for key, value in [("description", description), ("units", units), ("stepped", stepped)]:
+        for key, value in [
+            ("description", description),
+            ("units", units),
+            ("stepped", stepped),
+        ]:
             if value is not None:
                 metadata[key] = value
 
-        body = {
-            "entries": {
-                f"{tag}": metadata
-            },
-            "updateIfExists": update_existing
-        }
+        body = {"entries": {f"{tag}": metadata}, "updateIfExists": update_existing}
+
+        createResp = _do_historian_multi_request(url, body, self.auth)
 
-        createResp = _do_historian_multi_request(url,body,self.auth)
-        
         if len(createResp["errors"].keys()):
             createSuccess = False
         else:
             createSuccess = True
-            
-        body = {
-            "points": {
-                tag: force_list(points)
-            },
-            "createTags": False
-        }
 
-        message = _do_historian_multi_request(self.writeurl,body,self.auth,True)
+        body = {"points": {tag: force_list(points)}, "createTags": False}
 
-        if message == 'All points written!':
+        message = _do_historian_multi_request(self.writeurl, body, self.auth, True)
+
+        if message == "All points written!":
             writeSuccess = True
         else:
             ret = writeSuccess = True
 
-        return {"success": createSuccess and writeSuccess, "createTag": {"success": createSuccess, "errors": createResp["errors"]}, "writePoints": {"success": writeSuccess, "message": message}}
+        return {
+            "success": createSuccess and writeSuccess,
+            "createTag": {"success": createSuccess, "errors": createResp["errors"]},
+            "writePoints": {"success": writeSuccess, "message": message},
+        }
 
+    # SECTION - Write Requests
 
-    #SECTION - Write Requests
-
-    
-    def writePoints(self, tag:str, points: Union[dict, list]):
+    def writePoints(self, tag: str, points: Union[dict, list]):
         """Write one or more points to a single tag
 
         Args:
@@ -527,32 +582,26 @@ class HistorianMulti (object):
 
         for point in points:
             if "status" not in point.keys():
-                point['status'] = "OK"
-            point['timestamp'] = time_to_epoch_millis(point['timestamp'])
-        
-        body = {
-            "points": {
-                tag: points
-            },
-            "createTags": True
-        }
+                point["status"] = "OK"
+            point["timestamp"] = time_to_epoch_millis(point["timestamp"])
 
-        message = _do_historian_multi_request(self.writeurl,body,self.auth,True)
+        body = {"points": {tag: points}, "createTags": True}
 
-        if message == 'All points written!':
+        message = _do_historian_multi_request(self.writeurl, body, self.auth, True)
+
+        if message == "All points written!":
             ret = {"success": True, "message": message}
         else:
             ret = {"success": False, "message": message}
 
         return ret
 
-
-    def writePointsBatch(self, data:Union[dict,list]):
+    def writePointsBatch(self, data: Union[dict, list]):
         """Write one or more points to multiple tags
 
         Args:
-            payloads: A dict (or list of dicts) of the form {tag1: [{"value": x, "timestamp": y}, ...], tag2: [...], ...}    
-        
+            payloads: A dict (or list of dicts) of the form {tag1: [{"value": x, "timestamp": y}, ...], tag2: [...], ...}
+
         Returns:
             None
 
@@ -577,58 +626,65 @@ class HistorianMulti (object):
         for key in payload.keys():
             for datapoint in payload[key]:
                 if "status" not in datapoint.keys():
-                    datapoint['status'] = "OK"
-                datapoint['timestamp'] = time_to_epoch_millis(datapoint['timestamp'])
-        
+                    datapoint["status"] = "OK"
+                datapoint["timestamp"] = time_to_epoch_millis(datapoint["timestamp"])
+
         payload = update_keys(payload, self.historian)
 
-        body = {
-            "points": payload,
-            "createTags": True
-        }
+        body = {"points": payload, "createTags": True}
 
-        message = _do_historian_multi_request(self.writeurl,body,self.auth,True)
+        message = _do_historian_multi_request(self.writeurl, body, self.auth, True)
 
-        if message == 'All points written!':
+        if message == "All points written!":
             ret = {"success": True, "message": message}
         else:
             ret = {"success": False, "message": message}
 
         return ret
 
-#SECTION - Utility Functions
 
-def createPoint(value:Union[int,float],timestamp:Union[int,str,datetime],status:str="OK"):
+# SECTION - Utility Functions
+
+
+def createPoint(
+    value: Union[int, float], timestamp: Union[int, str, datetime], status: str = "OK"
+):
     """Quickly Create a dict of the format required to push data to historians
 
-        Args:
-            value: The value to be pushed
-            timestamp: The timestamp to assign the value to. Accepts formats of epoch milliseconds, python datetimes, and iso format datetime strings
-            status: (Optional) Indicates the quality of the data, accepts "OK" or "BAD". Defaults to "OK"
+    Args:
+        value: The value to be pushed
+        timestamp: The timestamp to assign the value to. Accepts formats of epoch milliseconds, python datetimes, and iso format datetime strings
+        status: (Optional) Indicates the quality of the data, accepts "OK" or "BAD". Defaults to "OK"
 
-        Returns:
-            A structured dict for use in the WriteToTag method
-        """
-    if status not in ["OK","BAD"]:
+    Returns:
+        A structured dict for use in the WriteToTag method
+    """
+    if status not in ["OK", "BAD"]:
         raise EigenException("Bad Status Parameter, must be 'OK' or 'BAD'")
     return {"value": value, "timestamp": timestamp, "status": status}
 
-def get_historian_multi(eigenserver = None, default_historian = None) -> HistorianMulti:
+
+def get_historian_multi(
+    eigenserver=None, default_historian=None, timestamp_format="iso"
+) -> HistorianMulti:
     """Instantiate a historian object for the given instance.
 
     Args:
         eigenserver (Optional): An EigenServer Object linked to the ingenuity url containing the historian. Can be omitted if environmental variable "EIGENSERVER" exists and is equal to the Ingenuity base url
         default_historian (Optional): The name of the historian to be used if none is passed with a tag. Defaults to None, and will not hanbdle tags without historians.
-
+        timestamp_format (Optional): The format in which timestamps are returned. Accepts "iso" for ISO 8601 strings, or "epoch" for epoch milliseconds. Defaults to "iso".
     Returns:
         An Object that can be used to query historian data.
 
     """
     if eigenserver is None:
         eigenserver = get_default_server()
-    elif type(eigenserver) == str:
+    elif isinstance(eigenserver, str):
         eigenserver = EigenServer(eigenserver)
 
-    return HistorianMulti(eigenserver.getEigenServerUrl() + "historian" + "/", eigenserver.auth, default_historian)
-
-
+    return HistorianMulti(
+        eigenserver.getEigenServerUrl() + "historian" + "/",
+        eigenserver.auth,
+        default_historian,
+        timestamp_format,
+    )
