@@ -19,6 +19,9 @@ class RelationshipFileProcessor:
         self.add_system_properties = not config.get_no_system_properties()
 
         self.required_labels = config.get_required_labels()[section]
+        self.required_properties = config.get_required_relationship_properties_dict()[section]
+        self.unrequired_properties = config.get_unrequired_relationship_properties_list()[section]
+        self.required_relationships = config.get_required_relationships_dict()[section]
         self.creation_time_property = config.get_creation_time_property()[section]
         self.update_time_property = config.get_update_time_property()[section]
 
@@ -31,6 +34,7 @@ class RelationshipFileProcessor:
         self.validate_only = config.in_validation_mode()
         self.summarise = config.in_summary_mode()
 
+        self.blank_property_override = config.get_blank_property_override()
         self.blank_property_list = config.get_blank_property_list()
         self.batch_rate = config.get_batch_rate()
         self.start_from = config.get_start_from()
@@ -75,14 +79,22 @@ class RelationshipFileProcessor:
         last_row = self.get_last_row(num_data_rows)
         while this_row < last_row:
 
-            relationships = rm.validate_relationships(self.contents.get_column_values_list(this_row, relation_columns))
+            combined_relationships = {**self.required_relationships, **self.contents.get_column_values_dict(this_row, relation_columns)}
+            relationships = rm.validate_relationships(list(combined_relationships.values()))
+
+            row_properties = self.contents.get_property_values_dict(this_row, property_columns, self.blank_property_list, blank_override=self.blank_property_override)
+            properties = pm.combine_properties(self.required_properties, row_properties)
+            for unwanted in self.unrequired_properties:
+                temp = properties.pop(unwanted, None)
+
+            # Split out those properties that are dependent on others, so that we can add them in the right order
+            explicit_properties, calculated_properties = pm.split_out_functional_properties(properties)
 
-            properties = self.contents.get_property_values_dict(this_row, property_columns, self.blank_property_list)
             # If there are multiple From or To nodes, create a relationship query for each combination
-            for from_node in self.contents.get_property_values_list(this_row, from_columns, self.primary_property, self.primary_property):
+            for from_node in self.contents.get_property_values_list(this_row, from_columns, self.primary_property, self.primary_property, override_format='str'):
                 validated_csv_from_labels = lm.validate_labels(self.contents.get_column_values_list(this_row, from_label_columns))
                 from_labels = lm.sort_relationship_labels(self.required_labels, validated_csv_from_labels)
-                for to_node in self.contents.get_property_values_list(this_row, to_columns, self.primary_property, self.primary_property):
+                for to_node in self.contents.get_property_values_list(this_row, to_columns, self.primary_property, self.primary_property, override_format='str'):
                     validated_csv_to_labels = lm.validate_labels(self.contents.get_column_values_list(this_row, to_label_columns))
                     to_labels = lm.sort_relationship_labels(self.required_labels, validated_csv_to_labels)
 
@@ -109,8 +121,11 @@ class RelationshipFileProcessor:
                                     update_time = relation_ref_dot + self.update_time_property + '=datetime("' + time_now + '")'
                                     query_text = query_text.on_create().set_literal(create_time).set_literal(update_time)
 
-                                if len(properties) > 0:
-                                    query_text = query_text.set(properties=properties, property_ref=relation_ref_dot)
+                                if len(explicit_properties) > 0:
+                                    query_text = query_text.set(properties=explicit_properties, property_ref=relation_ref_dot)
+                                if len(calculated_properties) > 0:
+                                    query_text = query_text.set(properties=calculated_properties, property_ref=relation_ref_dot)
+
                                 query_text = query_text.toStr()
                                 query = Neo4jQuery('relationship', query_text, time_now)
 
@@ -163,13 +178,20 @@ class RelationshipFileProcessor:
         this_row = self.start_from-1
         last_row = self.get_last_row(num_data_rows)
         while this_row < last_row:
-            relationships = rm.validate_relationships(self.contents.get_column_values_list(this_row, relation_columns))
-            properties = self.contents.get_property_values_dict(this_row, property_columns, self.blank_property_list)
+
+            combined_relationships = {**self.required_relationships, **self.contents.get_column_values_dict(this_row, relation_columns)}
+            relationships = rm.validate_relationships(list(combined_relationships.values()))
+
+            row_properties = self.contents.get_property_values_dict(this_row, property_columns, self.blank_property_list, blank_override=self.blank_property_override)
+            properties = pm.combine_properties(self.required_properties, row_properties)
+            for unwanted in self.unrequired_properties:
+                temp = properties.pop(unwanted, None)
+
             # If there are multiple From or To nodes, create a relationship query for each combination
-            for from_node in self.contents.get_property_values_list(this_row, from_columns, self.primary_property, self.primary_property):
+            for from_node in self.contents.get_property_values_list(this_row, from_columns, self.primary_property, self.primary_property, override_format='str'):
                 validated_csv_from_labels = lm.validate_labels(self.contents.get_column_values_list(this_row, from_label_columns))
                 from_labels = lm.sort_relationship_labels(self.required_labels, validated_csv_from_labels)
-                for to_node in self.contents.get_property_values_list(this_row, to_columns, self.primary_property, self.primary_property):
+                for to_node in self.contents.get_property_values_list(this_row, to_columns, self.primary_property, self.primary_property, override_format='str'):
                     validated_csv_to_labels = lm.validate_labels(self.contents.get_column_values_list(this_row, to_label_columns))
                     to_labels = lm.sort_relationship_labels(self.required_labels, validated_csv_to_labels)
 
@@ -180,7 +202,7 @@ class RelationshipFileProcessor:
 
                                 explicit_properties, calculated_properties = pm.split_out_functional_properties(properties)
                                 new_labels = [from_labels, to_labels, relation, from_node.key, to_node.key, calculated_properties]
-                                query_data = [from_node.value, to_node.value, explicit_properties, calculated_properties, time_now]
+                                query_data = [from_node.value.replace('\\\\n', '\n'), to_node.value.replace('\\\\n', '\n'), explicit_properties, calculated_properties, time_now]
 
                                 # So now we know all about this particular update. Let's put it in the query dictionary
                                 try:
@@ -205,8 +227,8 @@ class RelationshipFileProcessor:
             from_labels = batch_keys[from_to_labels][0]
             to_labels = batch_keys[from_to_labels][1]
             relationship = batch_keys[from_to_labels][2]
-            from_key, from_type = pm.split_node_key(batch_keys[from_to_labels][3])
-            to_key, to_type = pm.split_node_key(batch_keys[from_to_labels][4])
+            from_node = batch_keys[from_to_labels][3]
+            to_node = batch_keys[from_to_labels][4]
             batch_num = 1
             while batch_num <= num_batches:
                 if batch_num == num_batches:
@@ -222,8 +244,8 @@ class RelationshipFileProcessor:
                     update = updates[(batch_num - 1) * self.batch_rate + this_size]
                     update_properties = update[2] or {}  # In case update_properties[2] is {}
                     func_properties = update[3] or {}  # In case update_properties[3] is {}
-                    update_properties[from_ref] = Neo4jProperty(from_ref, update[0], from_type)
-                    update_properties[to_ref] = Neo4jProperty(to_ref, update[1], to_type)
+                    update_properties[from_ref] = Neo4jProperty(from_ref, update[0], 'str')
+                    update_properties[to_ref] = Neo4jProperty(to_ref, update[1], 'str')
                     if self.add_system_properties:
                         update_properties[self.update_time_property] = Neo4jProperty(self.update_time_property, update[4])
                     payload.append(update_properties)
@@ -234,8 +256,8 @@ class RelationshipFileProcessor:
                 for a_payload in payload:
                     formatted_payload.append(property.format(a_payload))
                 # Build a MERGE query using queries
-                from_keys = " {" + from_key + ":" + payload_ref_dot + from_ref + "}"
-                to_keys = " {" + to_key + ":" + payload_ref_dot + to_ref + "}"
+                from_keys = " {" + from_node + ":" + payload_ref_dot + from_ref + "}"
+                to_keys = " {" + to_node + ":" + payload_ref_dot + to_ref + "}"
                 create_time = relation_ref_dot + self.creation_time_property + '=datetime("' + time_now + '")'
                 update_time = relation_ref_dot + self.update_time_property + '=datetime("' + time_now + '")'
 
